@@ -8,6 +8,7 @@
  * Copyright (c) 2004-2005 Yuval Fledel
  * Copyright (c) 2004-2007 Yura Pakhuchiy
  * Copyright (c)      2005 Cristian Klein
+ * Copyright (c) 2011-2020 Jean-Pierre Andre
  *
  * This utility will dump a file's attributes.
  *
@@ -118,6 +119,7 @@ static void version(void)
 	printf("    2003      Leonard Norrgård\n");
 	printf("    2004-2005 Yuval Fledel\n");
 	printf("    2004-2007 Yura Pakhuchiy\n");
+	printf("    2011-2018 Jean-Pierre Andre\n");
 	printf("\n%s\n%s%s\n", ntfs_gpl, ntfs_bugs, ntfs_home);
 }
 
@@ -302,7 +304,8 @@ static int parse_options(int argc, char *argv[])
 	if (help || err)
 		usage();
 
-	return (!err && !help && !ver);
+		/* tri-state 0 : done, 1 : error, -1 : proceed */
+	return (err ? 1 : (help || ver ? 0 : -1));
 }
 
 
@@ -328,7 +331,7 @@ static char *ntfsinfo_time_to_str(const sle64 sle_ntfs_clock)
 		    "Jul", "Aug", "Sep", "Oct", "Nov", "Dec" } ;
 	static const char *wdays[]
 		= { "Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat" } ;
-	static char str[30];
+	static char str[50];
 	long long stamp;
 	u32 days;
 	u32 seconds;
@@ -368,7 +371,7 @@ static char *ntfsinfo_time_to_str(const sle64 sle_ntfs_clock)
 		mon = days/31 + 1;
 		days -= 31*(mon - 1) - 1;
 	}
-	sprintf(str,"%3s %3s %2u %02u:%02u:%02u %4u UTC\n",
+	snprintf(str, sizeof(str), "%3s %3s %2u %02u:%02u:%02u %4u UTC\n",
 		wdays[wday],
 		months[mon-1],(unsigned int)days,
 		(unsigned int)(seconds/3600),
@@ -405,6 +408,58 @@ static char *ntfs_attr_get_name_mbs(ATTR_RECORD *attr)
 		return NULL;
 }
 
+static const char *reparse_type_name(le32 tag)
+{
+	const char *name;
+	le32 seltag;
+
+	seltag = tag & IO_REPARSE_PLUGIN_SELECT;
+	switch (seltag) {
+	case IO_REPARSE_TAG_MOUNT_POINT :
+		name = " (mount point)";
+		break;
+	case IO_REPARSE_TAG_SYMLINK :
+		name = " (symlink)";
+		break;
+	case IO_REPARSE_TAG_WOF :
+		name = " (Wof compressed)";
+		break;
+	case IO_REPARSE_TAG_DEDUP :
+		name = " (deduplicated)";
+		break;
+	case IO_REPARSE_TAG_WCI :
+		name = " (Windows container)";
+		break;
+	case IO_REPARSE_TAG_CLOUD :
+		name = " (Cloud)";
+		break;
+	case IO_REPARSE_TAG_NFS :
+		name = " (NFS symlink)";
+		break;
+	case IO_REPARSE_TAG_LX_SYMLINK :
+		name = " (Linux symlink)";
+		break;
+	case IO_REPARSE_TAG_LX_FIFO :
+		name = " (Linux fifo)";
+		break;
+	case IO_REPARSE_TAG_LX_CHR :
+		name = " (Linux character device)";
+		break;
+	case IO_REPARSE_TAG_LX_BLK :
+		name = " (Linux block device)";
+		break;
+	case IO_REPARSE_TAG_AF_UNIX :
+		name = " (Unix socket)";
+		break;
+	case IO_REPARSE_TAG_APPEXECLINK :
+		name = " (Exec link)";
+		break;
+	default :
+		name = "";
+		break;
+	}
+	return (name);
+}
 
 /* *************** functions for dumping global info ******************** */
 /**
@@ -417,7 +472,7 @@ static void ntfs_dump_volume(ntfs_volume *vol)
 	printf("\tDevice state: %lu\n", vol->dev->d_state);
 	printf("\tVolume Name: %s\n", vol->vol_name);
 	printf("\tVolume State: %lu\n", vol->state);
-	printf("\tVolume Flags: 0x%04x", (int)vol->flags);
+	printf("\tVolume Flags: 0x%04x", (int)le16_to_cpu(vol->flags));
 	if (vol->flags & VOLUME_IS_DIRTY)
 		printf(" DIRTY");
 	if (vol->flags & VOLUME_MODIFIED_BY_CHKDSK)
@@ -579,6 +634,10 @@ static void ntfs_dump_flags(const char *indent, ATTR_TYPES type, le32 flags)
 		printf(" VIEW_INDEX");
 		flags &= ~FILE_ATTR_VIEW_INDEX_PRESENT;
 	}
+	if (flags & FILE_ATTRIBUTE_RECALL_ON_OPEN) {
+		printf(" RECALL_ON_OPEN");
+		flags &= ~FILE_ATTRIBUTE_RECALL_ON_OPEN;
+	}
 	if (flags)
 		printf(" UNKNOWN: 0x%08x", (unsigned int)le32_to_cpu(flags));
 	/* Print all the flags in hex. */
@@ -773,6 +832,8 @@ static void ntfs_dump_attr_list(ATTR_RECORD *attr, ntfs_volume *vol)
 static void ntfs_dump_filename(const char *indent,
 		FILE_NAME_ATTR *file_name_attr)
 {
+	le32 tag;
+
 	printf("%sParent directory:\t %lld (0x%llx)\n", indent,
 			(long long)MREF_LE(file_name_attr->parent_directory),
 			(long long)MREF_LE(file_name_attr->parent_directory));
@@ -810,10 +871,12 @@ static void ntfs_dump_filename(const char *indent,
 			(unsigned)file_name_attr->file_name_length);
 	ntfs_dump_flags(indent, AT_FILE_NAME, file_name_attr->file_attributes);
 	if (file_name_attr->file_attributes & FILE_ATTR_REPARSE_POINT &&
-			file_name_attr->reparse_point_tag)
-		printf("%sReparse point tag:\t 0x%x\n", indent, (unsigned)
-				le32_to_cpu(file_name_attr->reparse_point_tag));
-	else if (file_name_attr->reparse_point_tag) {
+			file_name_attr->reparse_point_tag) {
+		tag = file_name_attr->reparse_point_tag;
+		printf("%sReparse point tag:\t 0x%08lx%s\n", indent,
+				(long)le32_to_cpu(tag),
+				reparse_type_name(tag));
+	} else if (file_name_attr->reparse_point_tag) {
 		printf("%sEA Length:\t\t %d (0x%x)\n", indent, (unsigned)
 				le16_to_cpu(file_name_attr->packed_ea_size),
 				(unsigned)
@@ -1428,6 +1491,7 @@ static void ntfs_dump_index_key(INDEX_ENTRY *entry, INDEX_ATTR_TYPE type)
 {
 	char *sid;
 	char printable_GUID[37];
+	le32 tag;
 
 	switch (type) {
 	case INDEX_ATTR_SECURE_SII:
@@ -1451,8 +1515,10 @@ static void ntfs_dump_index_key(INDEX_ENTRY *entry, INDEX_ATTR_TYPE type)
 		ntfs_log_verbose("\t\tKey GUID:\t\t %s\n", printable_GUID);
 		break;
 	case INDEX_ATTR_REPARSE_R:
-		ntfs_log_verbose("\t\tKey reparse tag:\t 0x%08x\n", (unsigned)
-				le32_to_cpu(entry->key.reparse.reparse_tag));
+		tag = entry->key.reparse.reparse_tag;
+		ntfs_log_verbose("\t\tKey reparse tag:\t 0x%08lx%s\n",
+				(long)le32_to_cpu(tag),
+				reparse_type_name(tag));
 		ntfs_log_verbose("\t\tKey file id:\t\t %llu (0x%llx)\n",
 				(unsigned long long)
 				le64_to_cpu(entry->key.reparse.file_id),
@@ -1692,7 +1758,7 @@ static INDEX_ATTR_TYPE get_index_attr_type(ntfs_inode *ni, ATTR_RECORD *attr,
 		else
 			/* weird, this should be illegal */
 			ntfs_log_error("Unknown index attribute type: 0x%0X\n",
-				       index_root->type);
+				       le32_to_cpu(index_root->type));
 		return INDEX_ATTR_UNKNOWN;
 	}
 
@@ -1943,9 +2009,49 @@ static void ntfs_dump_attr_bitmap(ATTR_RECORD *attr __attribute__((unused)))
  *
  * of ntfs 3.x dumps the reparse_point attribute
  */
-static void ntfs_dump_attr_reparse_point(ATTR_RECORD *attr __attribute__((unused)))
+static void ntfs_dump_attr_reparse_point(ATTR_RECORD *attr
+			__attribute__((unused)), ntfs_inode *inode)
 {
-	/* TODO */
+	REPARSE_POINT *reparse;
+	le32 tag;
+	const char *name;
+	u8 *pvalue;
+	s64 size;
+	unsigned int length;
+	unsigned int cnt;
+
+	if (attr->non_resident) {
+		reparse = ntfs_attr_readall(inode, AT_REPARSE_POINT,
+						(ntfschar*)NULL, 0, &size);
+	} else {
+		reparse = (REPARSE_POINT*)((u8*)attr +
+				le16_to_cpu(attr->value_offset));
+	}
+	if (reparse) {
+		tag = reparse->reparse_tag;
+		name = reparse_type_name(tag);
+		printf("\tReparse tag:\t\t 0x%08lx%s\n",
+			(long)le32_to_cpu(tag),name);
+		length = le16_to_cpu(reparse->reparse_data_length);
+		printf("\tData length:\t\t %u (0x%x)\n",
+			(unsigned int)length,(unsigned int)length);
+		cnt = length;
+		pvalue = reparse->reparse_data;
+		printf("\tData:\t\t\t");
+		printf(cnt ? " 0x" : "(NONE)");
+		if (cnt > 32)
+			cnt = 32;
+		while (cnt-- > 0)
+			printf("%02x",*pvalue++);
+		if (length > 32)
+			printf("...\n");
+		else
+			printf("\n");
+		if (attr->non_resident)
+			free(reparse);
+	} else {
+		ntfs_log_perror("Failed to get the reparse data");
+	}
 }
 
 /**
@@ -1977,9 +2083,12 @@ static void ntfs_dump_attr_ea_information(ATTR_RECORD *attr)
  */
 static void ntfs_dump_attr_ea(ATTR_RECORD *attr, ntfs_volume *vol)
 {
-	EA_ATTR *ea;
+	const EA_ATTR *ea;
+	const u8 *pvalue;
 	u8 *buf = NULL;
-	le32 *pval;
+	const le32 *pval;
+	int offset;
+	int cnt;
 	s64 data_size;
 
 	if (attr->non_resident) {
@@ -2017,6 +2126,7 @@ static void ntfs_dump_attr_ea(ATTR_RECORD *attr, ntfs_volume *vol)
 			return;
 		ea = (EA_ATTR*)((u8*)attr + le16_to_cpu(attr->value_offset));
 	}
+	offset = 0;
 	while (1) {
 		printf("\n\tEA flags:\t\t ");
 		if (ea->flags) {
@@ -2033,21 +2143,36 @@ static void ntfs_dump_attr_ea(ATTR_RECORD *attr, ntfs_volume *vol)
 		printf("\tValue length:\t %d (0x%x)\n",
 				(unsigned)le16_to_cpu(ea->value_length),
 				(unsigned)le16_to_cpu(ea->value_length));
+			/* Name expected to be null terminated ? */
 		printf("\tName:\t\t '%s'\n", ea->name);
 		printf("\tValue:\t\t ");
 		if (ea->name_length == 11 &&
 				!strncmp((const char*)"SETFILEBITS",
 				(const char*)ea->name, 11)) {
-			pval = (le32*)(ea->value + ea->name_length + 1);
+			pval = (const le32*)(ea->value + ea->name_length + 1);
 			printf("0%lo\n", (unsigned long)le32_to_cpu(*pval));
+		} else {
+			/* No alignment for value */
+			pvalue = ea->value + ea->name_length + 1;
+			/* Hex show a maximum of 32 bytes */
+			cnt = le16_to_cpu(ea->value_length);
+			printf(cnt ? "0x" : "(NONE)");
+			if (cnt > 32)
+				cnt = 32;
+			while (cnt-- > 0)
+				printf("%02x",*pvalue++);
+			if (le16_to_cpu(ea->value_length) > 32)
+				printf("...\n");
+			else
+				printf("\n");
+		}
+		if (ea->next_entry_offset) {
+			offset += le32_to_cpu(ea->next_entry_offset);
+			ea = (const EA_ATTR*)((const u8*)ea
+					+ le32_to_cpu(ea->next_entry_offset));
 		} else
-			printf("'%s'\n", ea->value + ea->name_length + 1);
-		if (ea->next_entry_offset)
-			ea = (EA_ATTR*)((u8*)ea +
-					le32_to_cpu(ea->next_entry_offset));
-		else
 			break;
-		if ((u8*)ea - buf >= data_size)
+		if (offset >= data_size)
 			break;
 	}
 	free(buf);
@@ -2278,7 +2403,7 @@ static void ntfs_dump_file_attributes(ntfs_inode *inode)
 			ntfs_dump_attr_bitmap(ctx->attr);
 			break;
 		case AT_REPARSE_POINT:
-			ntfs_dump_attr_reparse_point(ctx->attr);
+			ntfs_dump_attr_reparse_point(ctx->attr, inode);
 			break;
 		case AT_EA_INFORMATION:
 			ntfs_dump_attr_ea_information(ctx->attr);
@@ -2325,15 +2450,17 @@ static void ntfs_dump_file_attributes(ntfs_inode *inode)
 int main(int argc, char **argv)
 {
 	ntfs_volume *vol;
+	int res;
 
 	setlinebuf(stdout);
 
 	ntfs_log_set_handler(ntfs_log_handler_outerr);
 
-	if (!parse_options(argc, argv)) {
+	res = parse_options(argc, argv);
+	if (res > 0)
 		printf("Failed to parse command line options\n");
-		exit(1);
-	}
+	if (res >= 0)
+		exit(res);
 
 	utils_set_locale();
 
@@ -2355,8 +2482,20 @@ int main(int argc, char **argv)
 		ntfs_inode *inode;
 		/* obtain the inode */
 		if (opts.filename) {
+#ifdef HAVE_WINDOWS_H
+			char *unix_name;
+
+			unix_name = ntfs_utils_unix_path(opts.filename);
+			if (unix_name) {
+				inode = ntfs_pathname_to_inode(vol, NULL,
+						unix_name);
+				free(unix_name);
+			} else
+				inode = (ntfs_inode*)NULL;
+#else
 			inode = ntfs_pathname_to_inode(vol, NULL,
 					opts.filename);
+#endif
 		} else {
 			inode = ntfs_inode_open(vol, MK_MREF(opts.inode, 0));
 		}
